function [Ym,Yp] = WdpwBX(A, IC, Ix, Iy, xmaxlooporder,ymaxlooporder,isconstant)
%Updated 12 Nov 2014

%The same as WdpwB, with some subfunctions implemented in C++
%  - thanks to Peter Nrtoft
%Needs the mex files "loop2SMatC.mex64" and "rLoopEval.mex64"

%Computes solutions for the Bjorling problem for Willmore surfaces
%Important:
%  1) Assumes the potential is real-valued along y=0. 
%  2) For CONSTANT potentials (equivariant) use "isconstant =1"
%      - this will be many times faster
%  3) For all other potentials use "isconstant=0" - otherwise the 
%  wrong solution will result

%Computes the frame for a harmonic map in SO(1,4)/(SO(1,1)xSO(3))
%using the DPW (generalized Weierstrass) method
%from  a Bjorling type of potential A.  
%The potential is entered twisted, and in the appropriate form for the 
%Bjorling problem for Willmore surfaces.
% A pair of dual surfaces [Ym, Yp] are computed, and Ym is plotted,
% stereographically projected by default from (0,0,0,1).
% The curve y=0 is plotted in red

%The computation is done over a rectangular grid,
%determined by Ix=[x_0, stepsize, pointsleft, pointsright]
%and Iy of similar form. 
%Loops are aproximated by polynomials of orders depending on 'xmaxlooporder'
% and 'ymaxlooporder'.  If the rectangle is longer in the x direction, then
% 'xmaxlooporder' should be larger, and vice versa.  
%The maximum errors (obtained by measuring how far the matrices are from
%being in SO(1,4))  are output. Normally if max x errors is higher than 10^(-2),
%then you should increase xmaxlooporder, and similarly with max y errors.  

%Authors: Matlab code: David Brander;  email: dbra@dtu.dk
%C++ re-implementations by Peter Nrtoft

% A = @(z) [A_-n(z),... A_n(z)]
% where each A_i(z,h) is an m x m matrix valued expression  holomorphic in z.
%This represents the potential (A_-n lambda^{-1} + A_1 lambda + ... + A_n
%lambda^n) dz. 

% All loops are converted into large square matrices, such that every 5-tuple of rows is 
%a copy of the previous 5-tuple, shifted to the right by m columns. Then matrix
% multiplication corresponds exactly to loop multiplication, and the LU
% decomposition corresponds to the birkhoff decomposition of loops.

%IC  represenst the initial condition, and is a Laurent
%polynomial loop: IC = [F_-j , ...., F_j]
%    where F_i  are constant mxm  matrices.  Usually IC = eye(5).

%Ix should be of the form [x_0, stepsize, n1, n2],  where n1 is the number
%of steps to the left of x_0 and n2, the number to the right.  Similar for
%Iy.

%Set "isconstant=1" if the potential is a constant function handle

%OUTPUT:
%Ym and Yp are the data for a pair of dual Willmore surfaces, lifted to 
%the light cone.  They need to be projected. Use the function 'projplot'.

%bcdistance should be zero.  

lambda=1;                     %in principle lambda could be any complex number of modulus 1.
minlooporder=4;               %The lowest loop order that will be used in the computations.
potorder = (length(A(Ix(1)+1i*Iy(1)))-5)/10;     %the looporder of the potential A
maxlooporder=max(ymaxlooporder,potorder);        %The highest loop order that will be used in the computations. 
intlooporder=max(ceil(maxlooporder/2)+1,potorder); %the looporder used for the integration in the y direction
                             %this only needs to be about half of "maxlooporder"
                             %because after the integration two loops will be 
                             %multiplied together (doubling their polynomial degree)
xintlooporder=max(ceil(xmaxlooporder/2)+1,potorder);   %the looporder used for the integration in the x direction
ICsize=size(IC);         
m=ICsize(1);                  %the dimension of the finite dimensional matrices
xpoints=1+Ix(3)+Ix(4);        %the number of x points in the grid
ypoints=1+Iy(3)+Iy(4);

bigICx=loop2SMatC(eye(m),xintlooporder,xintlooporder);    %The computations are
                                  %done with the identity as initial
                                  %condition. The final answer will be
                                  %premultiplied by the value of "IC".
bigIC=loop2SMatC(eye(m),intlooporder,intlooporder);

if isconstant                                %this option saves computation time in the integration,
   Amat=loop2SMatC(A(0),intlooporder,intlooporder); %by avoiding future calls to loop2matrix
   bigA=@(z)Amat;                       %the potential along the y axis in matrix form
   Amatx=loop2SMatC(A(0),xintlooporder,xintlooporder);
   bigAx=@(z)Amatx;                     %the potential along the x axis 
else
   bigA=@(t)loop2SMatC(A(t),intlooporder,potorder);     %the potential along the y axis
   bigAx=@(t)loop2SMatC(A(t),xintlooporder,potorder);   %the potential along the x axis
end

ICval=rLoopEval(IC,lambda);  %the initial condition evaluated at lambda=1
                               %will be used to premultiply the solution
%first integrate along the line x axis:  
[F1,F1errors]=realSintegrateA(bigAx, bigICx, Ix,m,lambda);
bsize=m*(2*intlooporder+1);  %length(bigIC), the size of the smaller big matrices
                             %representing the loops
                             
%then integrate along each line x=x_i:
bcdistance=0;              %bcdistance is increased whenever one of the lu decompositions
                           %fails to be in the big cell.  
xline = makexline(Ix);     %The set of gridpoints along the x-axis
looporders = spreadlooporders(Iy,minlooporder,maxlooporder);   %looporders(i,j) is the order of approximation
                                                   %used for loops in the lu-decomposition at the
                                                   %gridpoint (i,j).
disp(['x errors: ', num2str( F1errors',1)]);  
F12=cell(xpoints,1);         %this is the solution matrix, only the first two columns
                             % of the calculated frame F are needed.
    Maxerrors=zeros(xpoints,1);
    Meanerrors=zeros(xpoints,1);   %measures how much the first two columns of F deviated from being in SO(1,4,C)
    temp=sparse(zeros(5,2*ypoints)); % the solutions along the line x=xline(row) 
for row=1:xpoints  
    if (isconstant==0)||(row==1) %only does this once for constant potentials
        [Phi,Phiinv]=imagSintegrateA(bigA,bigIC, Iy, xline(row),m); 
        SOerror=zeros(ypoints,1);
        f1f2error=zeros(ypoints,1);
        for j=1:ypoints
            n=looporders(j);              %these three lines are just setting 
                                          %loop orders to speed things up.
            k=min([ceil(n/2)+2,intlooporder,n]);
            k2=max([2,k]);   %k2 is the order used in the LU
                      %we need to compute the lu decomp of conj(Phi)^(-1)* Phi
                      %The next three lines get an expression for conj(Phi)^(-1)* Phi
                      %from Phinv (which is a loop in column form) and Phi, which is 
                      %a loop in row form.
            thisPhi=loop2SMatC(Phi(:, (j-1)*bsize+1:j*bsize),n,k);
            thisPhiflipinv=loopflip2(conj(Phiinv(:,(j-1)*m+1:j*m)),m,n);
            tosplit=loop2SMatC(thisPhiflipinv*thisPhi,k2,k2);
            [Uval,p]=luid(tosplit,m,0.00000001,lambda,k2);      %a normalized lu decomposition
            bcdistance=bcdistance+p;           %bcdistance increases if P is not in the big cell
            Fval=rLoopEval(Phi(:, (j-1)*bsize+1:j*bsize),lambda)/Uval;
            J=[-1,0,0,0,0;0,1,0,0,0;0,0,1,0,0;0,0,0,1,0;0,0,0,0,1];
            SOerror(j)=compare(Fval.'*J*Fval,J); 
            f1f2error(j)=compare(Fval(:,1:2).'*J*Fval(:,1:2),J(1:2,1:2));
            temp(:,2*(j-1)+1:2*j)=Fval(:,1:2);
            clear thisPhi; clear thisPhiflipinv; clear tosplit;
        end
        Maxerrors(row)=max(f1f2error);
        Meanerrors(row)=mean(f1f2error,1);
        if mod(row,10)==1
             disp(['Row ', num2str(row),'.  SO check:  ', num2str( Maxerrors(row)), ', ', num2str(Meanerrors(row))]); 
             disp(['y errors: ', num2str( f1f2error',1)]);
        end
    end
    thisinit=ICval*F1(:,(row-1)*m+1:row*m);
    F12{row}=thisinit*temp;   
end
disp(['Max x error: ', num2str(max(F1errors),1),', Max y Error=', num2str(max(Maxerrors)), ', Mean y Error=', num2str(mean(Meanerrors)),', BC distance=', num2str(bcdistance)])
[Ym,Yp]=convertframe(F12);      %If F=[f1,...,f5], then Ym= f1-f2 and Yp=f1+f2. 
                                %These are the pair of Willmores surfaces (in
                                %the light-cone of Minkowski 5-space).                              
  
projplot(Ym,4,Iy(3)+1);  %Plot a projection of Ym
end



   
%%%%%%%%   CRITICAL SUB-FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
%%%%%%%%% The following 5 functions are called many times
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   function [F,Finv] = imagSintegrateA(A, F0, Iy, xvalue,m)
%The integration along the y axis.
%solves the equation dF/dt = F A(xvalue + i*t), with initial condition F(t0) = F0, with t
%in the specified interval
%The inverse matrix of F is also obtained, by solving dG/dt = -A(xvalue + i*t) G
%(This is faster than inverting the matrix later)

%The values of F are returned in loops in row form
%The vlaues of Finv are returned as loops in column form.
tempIy3=Iy(3); tempIy4=Iy(4);
if Iy(3)<2, tempIy3=2; end
if Iy(4)<2, tempIy4=2; end
stepy= Iy(2);   intervalyr=zeros(1,tempIy4+1); intervalyl=zeros(1,tempIy3+1);
for j = 1 : tempIy4+1
    intervalyr(j)=  Iy(1) + (j-1)*stepy;
end
for j = 1 : tempIy3+1
    intervalyl(j)=  Iy(1) - (j-1)*stepy;
end

bsize=length(F0);
n =(bsize/m-1)/2; % the order of X in lambda
initF=reshape(F0(m*n+1:m*n+m,:),m*bsize,1);
initF2=reshape(F0(:, m*n+1:m*n+m),m*bsize,1);
testfn2 = @(t,y) reshape((-1i*A(xvalue+1i*t))*sparse(reshape(y,bsize,m)), m*bsize,1);
testfn = @(t,y) reshape(sparse(reshape(y,m,bsize))*(1i*A(xvalue+1i*t)), m*bsize,1);
if Iy(3)>0
  [~,soll] = ode45(testfn ,intervalyl,initF);
  F1=sparse(reshape(flipud(soll(2:Iy(3)+1, :)).', m,bsize*(Iy(3))));
  [~,soll2] = ode45(testfn2 ,intervalyl,initF2);
  G1=sparse(reshape(flipud(soll2(2:Iy(3)+1, :)).', bsize,m*(Iy(3))));
else
    F1=[]; G1=[];
end
if Iy(4)>0
  [~,solr] = ode45(testfn ,intervalyr,initF);
  F2=sparse(reshape(solr(2:Iy(4)+1, :).', m,bsize*(Iy(4))));
  [~,solr2] = ode45(testfn2 ,intervalyr,initF2);
  G2=sparse(reshape(solr2(2:Iy(4)+1, :).', bsize,m*(Iy(4))));
else
    F2=[]; G2=[];
end
F=sparse([F1, F0(m*n+1:m*n+m,:) F2]);
Finv=[G1, F0(:, m*n+1:m*n+m), G2 ];
   end

        
         
 function [V,p]=luid(X,m,thresh,r,n)
   %computes a normalized lu decomposition of X. The middle block diagonal
   %of m x m blocks is  normalized to the identity.
   % returns the normalized matrix U evaluated at lambda = r, as V
   %P is 0 if X is in the big cell
[Lmat,Umat,P]=lu(X(1:m*(n+1),1:m*(2*n+1)),thresh);
p=trace(P)-length(P);
UmDims=size(Umat); LmDims=size(Lmat);
Uloop= Lmat(LmDims(1)-m+1:LmDims(1),LmDims(2)-m+1:LmDims(2))*Umat(UmDims(1)-m+1:UmDims(1), m*n+1 :m*(2*n+1));
   %the loop U is multiplied by the constant term of the loop L
V=rLoopEval(sparse([zeros(m,m*n),Uloop]),r);

 end
 

function V=loopflip2(U,m,n)
    %Converts a loop in column form to one in row form.
    %V(lambda)=U(1/lambda)
    %input: U is a loop given as a block colum matrix
    % U= [U_(n1);  ...  ; U_(-n1)]
    % output:  V is in row form V= [V_(-n),  ... V_n], resized to order n,
    %  
    K=length(U)/m;  %the number of coefficients of U
    V=sparse(zeros(m,m*(2*n+1)));
    lendiff= (m*(2*n+1)-length(U))/2;
    if lendiff>=0
      for j=0:K-1
         V(:,lendiff+m*j+1:lendiff+m*j+m)=U(m*j+1:m*j+m, :);
      end
    else
      for j=0:2*n
         V(:,m*j+1:m*j+m)=U(-lendiff+m*j+1:-lendiff+m*j+m, :);
      end  
    end       
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%








%%%%%%%%%%%%%%%%%%%%%%%%  Functions that are only called once       ****
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    function [F,Error] = realSintegrateA(A, F0, Ix,m,lambda)
%solves the equation dF/dt = F A, with initial condition F(t0) = F0, with t
%in the specified interval.  
% Estimates the "error" at each point by checking how far the solution
% matrix is from being in SO(1,4). Returns the errors in a column matrix.
% (It should be enough to do this just at the endpoints actually).

tempIx3=Ix(3); tempIx4=Ix(4);
if Ix(3)<2, tempIx3=2; end
if Ix(4)<2, tempIx4=2; end

stepy= Ix(2);  intervalyl=zeros(1, tempIx3+1); intervalyr=zeros(1,tempIx4+1); 
for count = 1 : tempIx4+1
    intervalyr(count)=  Ix(1) + (count-1)*stepy;
end
for count=1: tempIx3+1
    intervalyl(count)=  Ix(1) - (count-1)*stepy;
end
bsize=length(F0);
n =(bsize/m-1)/2; % the order of X in lambda
initF=reshape(F0(m*n+1:m*n+m,:),m*bsize,1);
testfn = @(t,y) reshape(reshape(y,m,bsize)*A(t), m*bsize,1);
[~,sol1] = ode45(testfn ,intervalyl,initF); 
[~,sol2] = ode45(testfn ,intervalyr,initF);  
Floops=[reshape(flipud(sol1(2:Ix(3)+1, :)).', m,bsize*(Ix(3))),   reshape(sol2(1:Ix(4)+1, :).', m,bsize*(Ix(4)+1))];
F=zeros(m,m*(Ix(3)+Ix(4)+1));
Error=zeros(Ix(3)+Ix(4)+1,1);
J=[-1,0,0,0,0;0,1,0,0,0;0,0,1,0,0;0,0,0,1,0;0,0,0,0,1];
for j=0:Ix(3)+Ix(4)
    F(:,m*j+1:m*j+m)=rLoopEval(Floops(:, j*bsize+1:bsize*(j+1)),lambda);
    Error(j+1)=compare(F(:,m*j+1:m*j+m).'*J*F(:,m*j+1:m*j+m),J); 
end
    end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%  MINOR FUNCTIONS:  The following functions are only called  once, and
%%%%%  do not use much time.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   
     function orders = spreadlooporders(Iy,minlooporder,maxlooporder) 
     %Assigns the order of the loops to be used in the processing the
     % the data at different points. The highest orders are at the corners
     % of the grid.
orders=zeros(1,1+Iy(3)+Iy(4));
m=max(Iy(3),Iy(4));
order=zeros(1,m+1);
if m==1
    order(2)=maxlooporder;
else
    for count=1:m
       order(count+1)=floor((m-count)*minlooporder/(m-1)+(count-1)*maxlooporder/(m-1));
    end
end
orders(1:Iy(3)+1)=fliplr(order(1:Iy(3)+1));
orders(Iy(3)+1+1:Iy(3)+1+Iy(4))=order(2:Iy(4)+1);
     end

%%%%%%%%%%%%%%%%%%%%%%%%%%
   
     function xline = makexline(Ix) 
     %makes the interval of integration along the x axis,
xline=zeros(1,1+Ix(3)+Ix(4));
xline(Ix(3)+1) = Ix(1);
for count=1:Ix(3)
    xline(count)=Ix(1)-Ix(2)*(Ix(3)-count+1);
end
for count=1:Ix(4)
    xline(Ix(3)+1+count)=Ix(1)+Ix(2)*count;
end
     end

%%%%%%%%%%%%%%%%%%%%%%%%%%




function y = compare(X,Y)    %This is used for estimating errors
%computes the size of the difference between X and Y
%compared with the size of the largest of  X and Y.
y=norm(X-Y, 'fro')/(max(norm(X, 'fro'), norm(Y, 'fro')));
end              
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%% Extracting and plotting the surfaces from the frame:
function [Ym,Yp]=convertframe(F)
Fmat=cell2mat(F);
Fdims=size(Fmat); xpoints=Fdims(1)/5; ypoints=Fdims(2)/2;
Ym=zeros(5*xpoints,ypoints);
Yp=zeros(5*xpoints,ypoints);
  for j=1:ypoints
    Ym(:,j:j)=real(Fmat(:,2*j-1:2*j-1)-Fmat(:,2*j:2*j));
    Yp(:,j:j)=real(Fmat(:,2*j-1:2*j-1)+Fmat(:,2*j:2*j));   
  end 
end

function g = projplot(Y,k,curve)
%The projection to R^3 of Y=(Y0,Y1,Y2,Y3,Y4) given by 
%(Y1,...[ skip Y|k| ], ...,Y4)/(Y0-sign(k)*Y|k|)
%If Y0/Yk is positive then it's the projection to the Poincare ball
%e.g. if k=1 the formula is (Y2,Y3,Y4)/(Y0-Y1).
%
%curve is an integer n. plots the curve given by the n'th column of Y in a
%different color.  
Ydims=size(Y); xpoints=Ydims(1)/5; ypoints=Ydims(2);  
g=zeros(xpoints*3,ypoints);
ind=[1 2 3]; K=abs(k);
ind(K:3)=ind(K:3)+1;
for row=1:xpoints
  for j=1:ypoints
      g(row,j)=Y(5*(row-1)+1+ind(1),j)/(Y(5*row-4,j)-sign(k)*Y(5*(row-1)+1+K,j));
      g(xpoints+row,j)=Y(5*(row-1)+1+ind(2),j)/(Y(5*row-4,j)-sign(k)*Y(5*(row-1)+1+K,j));
      g(2*xpoints+row,j)=Y(5*(row-1)+1+ind(3),j)/(Y(5*row-4,j)-sign(k)*Y(5*(row-1)+1+K,j));
  end
end  
surf(g(1:xpoints, :), g(xpoints+1: 2*xpoints, :),  g(2*xpoints+1: 3*xpoints, :),'Clipping','off','BackFaceLight','unlit','EdgeAlpha',0.3,'FaceAlpha',1,'LineWidth',1,'FaceLighting','gouraud','DiffuseStrength',0.8,'EdgeLighting','gouraud','EdgeColor','w','FaceColor',[0.7,0.75,0.9])
daspect([1 1 1])
hold all
plot3(g(1:xpoints, curve:curve), g(xpoints+1: 2*xpoints, curve:curve),  g(2*xpoints+1: 3*xpoints, curve:curve),'color',[0.7 0 0],'LineWidth',1.2)
hold off
end




