#include "mex.h"
#include "Eigen/Sparse"
#include <complex>
#include <vector>
#include <cmath>
#include "MatlabEigenUtils.h"
#include <iostream> // FIXME: just for debugging

using namespace std;
using namespace Eigen;

//-------------------------------------------------------------------
// This function prints a help message on the MATLAB prompt
//-------------------------------------------------------------------
void writeHelpMessage() {

  mexPrintf(
    "\n"
    "help for MEX function \'rLoopEval\':\n"
    "\n"
    "Usage:\n"
    "  B = rLoopEval( A, r )\n"
    "\n"
    "Input:\n"
    "\n"
    "  A: double, [M x N], real/complex, full/sparse.\n"
    "    The size [M x N] of the loop matrix must obey: N = (1 + 2*p) * M\n"
    "    for some p in {0,1,2,...}. Hence, we may picture A as:\n"
    "\n"
    "      [ A_{-p}  A_{-p+1}  ...  A_{0}  ...  A_{p-1}  A_{p} ],\n"
    "\n"
    "    where A_{i} has size [M x M] for i in {-p, ..., p}.\n"
    "\n"
    "  r: double, [1 x 1], real/complex.\n"
    "    Point in which to evaluate the loop.\n"
    "\n"
    "Output:\n"
    "\n"
    "  B: double, [M x  M], real/complex, sparse.\n"
    "    Value of the loop in the specified point.\n"
    "\n"
    "Authors:\n"
    "  David Brander (dbra@dtu.dk)\n"
    "  Peter Noertoft (penn@dtu.dk)\n"
    "\n"
  );
  return;

}

//-------------------------------------------------------------------
// This is the main "gateway" function
//-------------------------------------------------------------------
void mexFunction ( int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[]) {
   
    // Get input ------------------------------------------------------------
  
    // Check number of input and output arguments.
    if (nrhs==0) { // If no inputs are given, then write the help message
      writeHelpMessage();
      // fill in empty outputs if requested to avoid an error message like 
      // "One or more output arguments not assigned during call to "rLoopEval""
      for (mwIndex i=0; i<nlhs; ++i) { 
        plhs[i] = mxCreateDoubleMatrix(0,0,mxREAL);
      }
      return;
    }
    if (nrhs!=2) {
      mexErrMsgIdAndTxt("MATLAB:rLoopEval:invalidNumInputs",
          "Two input arguments are required.\n"
          "For help, type 'rLoopEval()'.");
    } else if (nlhs>1) {
      mexErrMsgIdAndTxt("MATLAB:rLoopEval:invalidNumOutputs",
          "No more than one output argument is allowed.\n"
          "For help, type 'rLoopEval()'.");
    } 
    
    // Get input: A
    SparseMatrix<complex<double> > A;
    try { 
        // Copy input MATLAB matrix to Eigen sparse matrix
        A = matrixPointer2SparseComplex(prhs[0]);
    } catch (exception& e) {
        string txt = "Incorrect value of input A.\n  --> ";
        txt       += string(e.what());
        txt       += string("\nFor help, type 'rLoopEval()'.");
        mexErrMsgIdAndTxt("MATLAB:rLoopEval:invalidInput:A", txt.c_str());
    }
    const bool isComplex = (mxGetPi(prhs[0]) != NULL);
    
    // Compute dimension of input matrix and check it.
    const size_t M = A.rows();
    const size_t N = A.cols();
    if ( (N % M)!=0 || ((N/M-1) % 2)!=0 ) {
      mexErrMsgIdAndTxt("MATLAB:loop2Matrix:invalidInput:A",
          "Incorrect dimension of input matrix A.\n"
          " --> If A is of size [M x N] then M and N must obey: "
          "N = (1 + 2*p) * M for given p in {0,1,2,...}"
          "For help, type 'loop2Matrix()'");
    }
    
    // Determine order of input loop from the size of the input matrix
    const int p = (N/M-1)/2;
    
    // Get input: r
    complex<double> r;
    try {
      r = mxArray2ScalarComplex<double>(prhs[1]);
    } catch (exception& e) {
      string txt = "Incorrect value of input r.\n  --> ";
      txt       += string(e.what());
      txt       += string("\nFor help, type 'rLoopEval()'.");
      mexErrMsgIdAndTxt("MATLAB:rLoopEval:invalidInput:r", txt.c_str());
    }
    
    // Fill vector of triplets from input matrix
    vector<ETD> triplets;
    triplets.reserve((2*p+1)*M*M); // assume the output matrix is full;
    for (int k = -p; k<=p; ++k ) {
        SparseMatrix<complex<double> > tmp(A.block(0,(k+p)*M,M,M));
        const complex<double> rToPowerK = pow(r,k);
        for (size_t l=0; l<(size_t)tmp.outerSize(); ++l) { 
            for (SparseMatrix<complex<double> >::InnerIterator it(tmp,l); it; ++it) {
                triplets.push_back(ETD(it.row(), it.col(), rToPowerK*it.value()));
            }
        }
    }
    
    // Construct Eigen matrix from vector of triplets. Note: Repeated entries are summed.
    SparseMatrix<complex<double> > B(M,M);
    B.setFromTriplets(triplets.begin(), triplets.end());
    
    // Construct MATLAB matrix from Eigen Matrix and return pointer to data.
    plhs[0] = sparseComplex2MatrixPointer(B, true);
    
}