#include "mex.h"
#include <string>
#include <vector>
#include <algorithm>
#include <complex>
#include "Eigen/Dense"
#include "Eigen/Sparse"
#include <algorithm>
#include <stdexcept>
#include <string>
#include "MatlabEigenUtils.h"
#include <iostream> // FIXME: just for debugging

using namespace std;
using namespace Eigen;

//-------------------------------------------------------------------
// This function prints a help message on the MATLAB prompt
//-------------------------------------------------------------------
void writeHelpMessage() {

  mexPrintf(
    "\n"
    "help for MEX function \'loop2Matrix\':\n"
    "\n"
    "Usage:\n"
    "  B = loop2Matrix( A, n, k )\n"
    "\n"
    "Input:\n"
    "\n"
    "  A: double, [M x N], real/complex, full/sparse.\n"
    "    The size [M x N] of the input matrix must obey: N = (1 + 2*p) * M\n"
    "    for some p in {0,1,2,...}. Hence, we may picture A as:\n"
    "\n"
    "      [ A_{-p}  A_{-p+1}  ...  A_{0}  ...  A_{p-1}  A_{p} ],\n"
    "\n"
    "    where A_{i} has size [M x M] for i in {-p, ..., p}.\n"
    "\n"
    "  n: double, [1 x 1], non-negative integer.\n"
    "    Number of copies of (subblocks of) A to make in B is\n"
    "    given by 4*n+1, see illustration of B below.\n"
    "\n"
    "  k: double, [1 x 1], non-negative integer.\n"
    "    Index of the subblocks of A to copy, i.e. copy\n"
    "\n"
    "      [ A_{-K} ... A_{0] ... A_{K} ],\n"
    "\n"
    "    where K = min( p, k ).\n"
    "\n"
    "Output:\n"
    "\n"
    "  B: double, [ M*(4*n+1) x  M*(4*n+1)], real/complex, sparse.\n"
    "    The output matrix contains copies of (subblocks of) the input\n"
    "    matrix. The structure/pattern of the output matrix is:\n"
    "\n"
    "      [  A_{ 0 }  A_{ 1  }    .     A_{ K }     0        .        .     ]  <1>\n"
    "      [  A_{-1 }  A_{ 0  }    .     A_{K-1}  A_{ K }     0        .     ]  <2>\n"
    "      [     .        .        .        .        .        .        .     ]\n"
    "      [  A_{-K }  A_{-K+1}    .        .        .        .        .     ]\n"
    "      [     0        .        .        .        .        .        .     ]\n"
    "      [     .        .        .        .        .        .        .     ]\n"
    "      [     .        .        .        .        .        .        0     ]\n"
    "      [     .        .        .        .        .     A_{K-1}  A_{ K }  ]\n"
    "      [     .        .        .        .        .        .        .     ]\n"
    "      [     .        0     A_{-K }  A_{-K+1}    .     A_{ 0 }  A_{ 1 }  ]  <4*n>\n"
    "      [     .        .        0     A_{-K  }    .     A_{-1 }  A_{ 0 }  ]  <4*n+1>\n"
    "\n"
    "           <1>      <2>                                <4*n>   <4*n+1>\n"
    "\n"
    "Authors:\n"
    "  David Brander (dbra@dtu.dk)\n"
    "  Peter Noertoft (penn@dtu.dk)\n"
    "\n"
  );
  return;

}

//-------------------------------------------------------------------
// This is the main "gateway" function
//-------------------------------------------------------------------
//
//  * Computations are done in complex numbers always, independent of
//    whether input is real or complex. Output, however, has the same
//    type as input.

void mexFunction ( int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[] ) {

  // Get input ------------------------------------------------------------
  
  // Check number of input and output arguments.
  if (nrhs==0) { // If no inputs are given, then write the help message
    writeHelpMessage();
    // fill in empty outputs if requested to avoid an error message like 
    // "One or more output arguments not assigned during call to "loop2Matrix""
    for (mwIndex i=0; i<nlhs; ++i) { 
      plhs[i] = mxCreateDoubleMatrix(0,0,mxREAL);
    }
    return;
  }
  if (nrhs!=3) {
    mexErrMsgIdAndTxt("MATLAB:loop2Matrix:invalidNumInputs",
        "Three input arguments are required.\n"
        "For help, type 'loop2Matrix()'.");
  } else if (nlhs>1) {
    mexErrMsgIdAndTxt("MATLAB:loop2Matrix:invalidNumOutputs",
        "No more than one output argument is allowed.\n"
        "For help, type 'loop2Matrix()'.");
  }

  // Get input 1: matrix
  SparseMatrix<complex<double> > A;
  try {
    A = matrixPointer2SparseComplex(prhs[0]);
  } catch (exception& e) {
    string txt = "Incorrect value of input A.\n  --> ";
    txt       += string(e.what());
    txt       += string("\nFor help, type 'loop2Matrix()'.");
    mexErrMsgIdAndTxt("MATLAB:loop2Matrix:invalidInput:A", txt.c_str());
  }
  const bool isComplex = (mxGetPi(prhs[0]) != NULL);

  // Compute dimension of input matrix and check it.
  const size_t M = A.rows();
  const size_t N = A.cols();
  if ( (N % M)!=0 || ((N/M-1) % 2)!=0 ) {
    mexErrMsgIdAndTxt("MATLAB:loop2Matrix:invalidInput:A",
        "Incorrect dimension of input matrix A.\n"
        " --> If A is of size [M x N] then M and N must obey: "
        "N = (1 + 2*p) * M for given p in {0,1,2,...}"
        "For help, type 'loop2Matrix()'");
  }
  
  // Get input 2: n
  int n;
  try {
    n = mxArray2ScalarReal<int>(prhs[1]);
  } catch (exception& e) {
    string txt = "Incorrect value of input n.\n  --> ";
    txt       += string(e.what());
    txt       += string("\nFor help, type 'loop2Matrix()'.");
    mexErrMsgIdAndTxt("MATLAB:loop2Matrix:invalidInput:n", txt.c_str());
  }
  if (n < 0) {
    mexErrMsgIdAndTxt("MATLAB:loop2Matrix:invalidInput",
        "Incorrect value of input n.\n"
        "  --> n must be a non-negative integer.\n"
        "For help, type 'loop2Matrix()'.");
  }

  // Get input 3: output order
  int k;
  try {
    k = mxArray2ScalarReal<int>(prhs[2]);
  } catch (exception& e) {
    string txt = "Incorrect value of input k.\n  --> ";
    txt       += string(e.what());
    txt       += string("\nFor help, type 'loop2Matrix()'.");
    mexErrMsgIdAndTxt("MATLAB:loop2Matrix:invalidInput:k", txt.c_str());
  }
  if (k < 0) {
    writeHelpMessage();
    mexErrMsgIdAndTxt("MATLAB:loop2Matrix:invalidInput:k",
        "Incorrect value of input k.\n"
        "  --> k must be a non-negative integer.\n"
        "For help, type 'loop2Matrix()'.");
  }

  // Reshape matrix -------------------------------------------------------

  // Determine output order.
  const int    p = (N/M-1)/2; // order of input loop
  const int    K = min( p , k ); // choose the smallest

  // Compute sum from 0 to K.
  int sumK = 0;
  for (int i=1; i<=K; ++i) {
    sumK+=i;
  }
  
  // Compute triplets (row,column,value) for the reshaped matrix
  std::vector<ETD> triplets;
  triplets.reserve(A.nonZeros());
  for (int k=-K; k<=K; ++k) {
    const size_t it0 = (k+K)*M,
                 jt0 = 0,
                 im0 = 0,
                 jm0 = ((2*p+1)-(k+p)-1)*M;
    Eigen::SparseMatrix<std::complex<double> > tmp(A.block(im0,jm0,M,M));
    for (int k=0; k<tmp.outerSize(); ++k) {
      for (Eigen::SparseMatrix<std::complex<double> >::InnerIterator it(tmp,k); it; ++it) {
        triplets.push_back( ETD(it.row()+it0, it.col()+jt0, it.value()) );
      }
    }
  }
  
  // Construct reshaped matrix from triplets
  Eigen::SparseMatrix<std::complex<double> > At((2*K+1)*M,M);
  At.setFromTriplets(triplets.begin(), triplets.end());
  
  // Prepare output matrix ------------------------------------------------
  
  // Compute dimensions of output matrix and the (maximal) number of non-zeroes.
  const int    numMat = 4*n+1;
  const size_t Nt     = M;
  const size_t Mo     = numMat * Nt;
  const size_t No     = numMat * Nt;
  const size_t nnzmax = ((2*K+1)*numMat - 2*sumK) * Nt * Nt;
  
  // Initiate output as a sparse, matrix and get pointers to the data.
  plhs[0] = mxCreateSparse(Mo, No, nnzmax, isComplex ? mxCOMPLEX : mxREAL);
  double  * pr = mxGetPr(plhs[0]);                    // real part
  double  * pi = isComplex ? mxGetPi(plhs[0]) : NULL; // imaginary part
  mwIndex * ir = mxGetIr(plhs[0]);                    // row number
  mwIndex * jc = mxGetJc(plhs[0]);                    // "counter" for column number
  
  // Fill output matrix ---------------------------------------------------
  
  // Fill in output matrix by copying suitable entries of the matrix
  *jc++        = 0;
  mwIndex inz  = 0;
  for (int k=0; k<numMat; ++k) {
    const size_t iBegin  = max(K-k,0)*Nt;
    const size_t iEnd    = min(numMat-k+K,2*K+1)*Nt;
    const size_t iOffset = max(k-K,0)*Nt;
    // Make a copy of the relevant block
    //SparseMatrix<complex<double> > tmp(At.block(iBegin,0,iEnd-iBegin,M));
    //for (int l=0; l<tmp.outerSize(); ++l) {
    //  for (SparseMatrix<complex<double> >::InnerIterator it(tmp,l); it; ++it) {
    //    *pr++ = it.value().real();
    //    if (isComplex) {
    //      *pi++ = it.value().imag();
    //    }	
    //    *ir++ = it.row()+iOffset;
    //    ++inz;
    //  }
    //  *jc++ = inz;
    //}
    // Loop over the entire block and check each element if it is relevant
    for (int l=0; l<At.outerSize(); ++l) {
      for (SparseMatrix<complex<double> >::InnerIterator it(At,l); it; ++it) {
        if (it.row()>=iBegin && it.row()<iEnd) {
            *pr++ = it.value().real();
            if (isComplex) {
              *pi++ = it.value().imag();
            }	
            *ir++ = it.row()-iBegin+iOffset;
            ++inz;
        }
      }
      *jc++ = inz;
    }
  }
  
  // That's it.
  return;

}
